"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var fast_glob_1 = __importDefault(require("fast-glob"));
var vscode_languageserver_1 = require("vscode-languageserver");
var util_1 = require("../common/util");
var logger_1 = __importDefault(require("../common/logger"));
var patterns_1 = require("../common/patterns");
var config_1 = __importDefault(require("./config"));
var builtin_docs_json_1 = __importDefault(require("../docs/builtin-docs.json"));
var log = logger_1.default('builtin');
var Builtin = /** @class */ (function () {
    function Builtin() {
        // completion items
        this.vimPredefinedVariablesItems = [];
        this.vimOptionItems = [];
        this.vimBuiltinFunctionItems = [];
        this.vimBuiltinFunctionMap = {};
        this.vimCommandItems = [];
        this.vimMapArgsItems = [];
        this.vimFeatureItems = [];
        this.vimAutocmdItems = [];
        this.expandKeywordItems = [];
        this.colorschemeItems = [];
        this.highlightArgKeys = [];
        this.highlightArgValues = {};
        // signature help
        this.vimBuiltFunctionSignatureHelp = {};
        // documents
        this.vimBuiltFunctionDocuments = {};
        this.vimOptionDocuments = {};
        this.vimPredefinedVariableDocuments = {};
        this.vimCommandDocuments = {};
        this.vimFeatureDocuments = {};
        this.expandKeywordDocuments = {};
    }
    Builtin.prototype.init = function () {
        this.start();
    };
    Builtin.prototype.getPredefinedVimVariables = function () {
        return this.vimPredefinedVariablesItems;
    };
    Builtin.prototype.getVimOptions = function () {
        return this.vimOptionItems;
    };
    Builtin.prototype.getBuiltinVimFunctions = function () {
        return this.vimBuiltinFunctionItems;
    };
    Builtin.prototype.isBuiltinFunction = function (label) {
        return this.vimBuiltinFunctionMap[label];
    };
    Builtin.prototype.getExpandKeywords = function () {
        return this.expandKeywordItems;
    };
    Builtin.prototype.getVimCommands = function () {
        return this.vimCommandItems;
    };
    Builtin.prototype.getVimMapArgs = function () {
        return this.vimMapArgsItems;
    };
    Builtin.prototype.getVimFeatures = function () {
        return this.vimFeatureItems;
    };
    Builtin.prototype.getVimAutocmds = function () {
        return this.vimAutocmdItems;
    };
    Builtin.prototype.getColorschemes = function () {
        return this.colorschemeItems;
    };
    Builtin.prototype.getHighlightArgKeys = function () {
        return this.highlightArgKeys;
    };
    Builtin.prototype.getHighlightArgValues = function () {
        return this.highlightArgValues;
    };
    Builtin.prototype.getSignatureHelpByName = function (name, idx) {
        var params = this.vimBuiltFunctionSignatureHelp[name];
        if (params) {
            return {
                signatures: [{
                        label: name + "(" + params[0] + ")" + (params[1] ? ": " + params[1] : ''),
                        documentation: this.formatVimDocument(this.vimBuiltFunctionDocuments[name]),
                        parameters: params[0].split('[')[0].split(',').map(function (param) {
                            return {
                                label: param.trim()
                            };
                        })
                    }],
                activeSignature: 0,
                activeParameter: idx
            };
        }
        return;
    };
    Builtin.prototype.getDocumentByCompletionItem = function (params) {
        var kind = params.kind;
        switch (kind) {
            case vscode_languageserver_1.CompletionItemKind.Variable:
                if (!this.vimPredefinedVariableDocuments[params.label]) {
                    return;
                }
                return __assign({}, params, { documentation: this.formatVimDocument(this.vimPredefinedVariableDocuments[params.label]) });
            case vscode_languageserver_1.CompletionItemKind.Property:
                if (!this.vimOptionDocuments[params.label]) {
                    return;
                }
                return __assign({}, params, { documentation: this.formatVimDocument(this.vimOptionDocuments[params.label]) });
            case vscode_languageserver_1.CompletionItemKind.Function:
                if (!this.vimBuiltFunctionDocuments[params.label]) {
                    return;
                }
                return __assign({}, params, { documentation: this.formatVimDocument(this.vimBuiltFunctionDocuments[params.label]) });
            case vscode_languageserver_1.CompletionItemKind.EnumMember:
                if (!this.vimFeatureDocuments[params.label]) {
                    return;
                }
                return __assign({}, params, { documentation: this.formatVimDocument(this.vimFeatureDocuments[params.label]) });
            case vscode_languageserver_1.CompletionItemKind.Operator:
                if (!this.vimCommandDocuments[params.label]) {
                    return;
                }
                return __assign({}, params, { documentation: this.formatVimDocument(this.vimCommandDocuments[params.label]) });
            default:
                break;
        }
    };
    Builtin.prototype.getHoverDocument = function (name, pre, next) {
        // builtin variables
        if (util_1.isSomeMatchPattern(patterns_1.builtinVariablePattern, pre) && this.vimPredefinedVariableDocuments[name]) {
            return {
                contents: this.formatVimDocument(this.vimPredefinedVariableDocuments[name])
            };
            // options
        }
        else if (util_1.isSomeMatchPattern(patterns_1.optionPattern, pre) && this.vimOptionDocuments[name.slice(1)]) {
            return {
                contents: this.formatVimDocument(this.vimOptionDocuments[name.slice(1)])
            };
            // builtin functions
        }
        else if (patterns_1.builtinFunctionPattern.test(next) && this.vimBuiltFunctionDocuments[name]) {
            return {
                contents: this.formatVimDocument(this.vimBuiltFunctionDocuments[name])
            };
            // has features
        }
        else if (util_1.isSomeMatchPattern(patterns_1.featurePattern, pre) && this.vimFeatureDocuments[name]) {
            return {
                contents: this.formatVimDocument(this.vimFeatureDocuments[name])
            };
            // expand Keywords
        }
        else if (util_1.isSomeMatchPattern(patterns_1.expandPattern, pre) && this.expandKeywordDocuments["<" + name + ">"]) {
            return {
                contents: this.formatVimDocument(this.expandKeywordDocuments["<" + name + ">"])
            };
            // command
        }
        if (util_1.isSomeMatchPattern(patterns_1.commandPattern, pre) && this.vimCommandDocuments[name]) {
            return {
                contents: this.formatVimDocument(this.vimCommandDocuments[name])
            };
        }
    };
    Builtin.prototype.start = function () {
        return __awaiter(this, void 0, void 0, function () {
            var runtimepath, data;
            var _this = this;
            return __generator(this, function (_a) {
                runtimepath = config_1.default.runtimepath;
                // get colorschemes
                if (runtimepath) {
                    this.resolveColorschemes(runtimepath);
                }
                // get map args
                this.resolveMapArgs();
                // get highlight arg keys
                this.resolveHighlightArgKeys();
                // get highlight arg values
                this.resolveHighlightArgValues();
                try {
                    data = builtin_docs_json_1.default;
                    this.vimBuiltinFunctionItems = data.completionItems.functions;
                    this.vimBuiltinFunctionItems.forEach(function (item) {
                        if (!_this.vimBuiltinFunctionMap[item.label]) {
                            _this.vimBuiltinFunctionMap[item.label] = true;
                        }
                    });
                    this.vimBuiltFunctionDocuments = data.documents.functions;
                    this.vimCommandItems = data.completionItems.commands;
                    this.vimCommandDocuments = data.documents.commands;
                    this.vimPredefinedVariablesItems = data.completionItems.variables;
                    this.vimPredefinedVariableDocuments = data.documents.variables;
                    this.vimOptionItems = data.completionItems.options;
                    this.vimOptionDocuments = data.documents.options;
                    this.vimFeatureItems = data.completionItems.features;
                    this.vimAutocmdItems = data.completionItems.autocmds;
                    this.vimFeatureDocuments = data.documents.features;
                    this.expandKeywordItems = data.completionItems.expandKeywords;
                    this.expandKeywordDocuments = data.documents.expandKeywords;
                    this.vimBuiltFunctionSignatureHelp = data.signatureHelp;
                }
                catch (error) {
                    log.error("[vimls]: parse docs/builtin-doc.json fail => " + (error.message || error));
                }
                return [2 /*return*/];
            });
        });
    };
    // format vim document to markdown
    Builtin.prototype.formatVimDocument = function (document) {
        var indent = 0;
        return {
            kind: vscode_languageserver_1.MarkupKind.Markdown,
            value: [
                '``` help'
            ].concat(document.map(function (line) {
                if (indent === 0) {
                    var m = line.match(/^([ \t]+)/);
                    if (m) {
                        indent = m[1].length;
                    }
                }
                return line.replace(new RegExp("^[ \\t]{" + indent + "}", 'g'), '').replace(/\t/g, '  ');
            }), [
                '```'
            ]).join('\n')
        };
    };
    Builtin.prototype.resolveMapArgs = function () {
        this.vimMapArgsItems = ["<buffer>", "<nowait>", "<silent>", "<script>", "<expr>", "<unique>"]
            .map(function (item) {
            return {
                label: item,
                kind: vscode_languageserver_1.CompletionItemKind.EnumMember,
                documentation: '',
                insertText: item,
                insertTextFormat: vscode_languageserver_1.InsertTextFormat.PlainText
            };
        });
    };
    Builtin.prototype.resolveColorschemes = function (runtimepath) {
        return __awaiter(this, void 0, void 0, function () {
            var list, glob, colorschemes, error_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        list = runtimepath;
                        if (config_1.default.vimruntime) {
                            list.push(config_1.default.vimruntime);
                        }
                        glob = runtimepath.map(function (p) { return path_1.default.join(p.trim(), 'colors/*.vim'); });
                        colorschemes = [];
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4 /*yield*/, fast_glob_1.default(glob, { onlyFiles: false, deep: 0 })];
                    case 2:
                        colorschemes = _a.sent();
                        return [3 /*break*/, 4];
                    case 3:
                        error_1 = _a.sent();
                        log.error([
                            "Index Colorschemes Error: " + JSON.stringify(glob),
                            "Error => " + (error_1.stack || error_1.message || error_1)
                        ].join('\n'));
                        return [3 /*break*/, 4];
                    case 4:
                        this.colorschemeItems = colorschemes.map(function (p) {
                            var label = path_1.default.basename(p, '.vim');
                            var item = {
                                label: label,
                                kind: vscode_languageserver_1.CompletionItemKind.EnumMember,
                                insertText: label,
                                insertTextFormat: vscode_languageserver_1.InsertTextFormat.PlainText
                            };
                            return item;
                        });
                        return [2 /*return*/];
                }
            });
        });
    };
    Builtin.prototype.resolveHighlightArgKeys = function () {
        this.highlightArgKeys = [
            'cterm',
            'start',
            'stop',
            'ctermfg',
            'ctermbg',
            'gui',
            'font',
            'guifg',
            'guibg',
            'guisp',
            'blend'
        ]
            .map(function (item) {
            return {
                label: item,
                kind: vscode_languageserver_1.CompletionItemKind.EnumMember,
                documentation: '',
                insertText: item + "=${0}",
                insertTextFormat: vscode_languageserver_1.InsertTextFormat.Snippet
            };
        });
    };
    Builtin.prototype.resolveHighlightArgValues = function () {
        var values = {
            'cterm': ['bold', 'underline', 'undercurl', 'reverse', 'inverse', 'italic', 'standout', 'NONE'],
            'ctermfg ctermbg': [
                'Black',
                'DarkBlue',
                'DarkGreen',
                'DarkCyan',
                'DarkRed',
                'DarkMagenta',
                'Brown', 'DarkYellow',
                'LightGray', 'LightGrey', 'Gray', 'Grey',
                'DarkGray', 'DarkGrey',
                'Blue', 'LightBlue',
                'Green', 'LightGreen',
                'Cyan', 'LightCyan',
                'Red', 'LightRed',
                'Magenta', 'LightMagenta',
                'Yellow', 'LightYellow',
                'White',
            ],
            'guifg guibg guisp': [
                'NONE',
                'bg',
                'background',
                'fg',
                'foreground',
                'Red', 'LightRed', 'DarkRed',
                'Green', 'LightGreen', 'DarkGreen', 'SeaGreen',
                'Blue', 'LightBlue', 'DarkBlue', 'SlateBlue',
                'Cyan', 'LightCyan', 'DarkCyan',
                'Magenta', 'LightMagenta', 'DarkMagenta',
                'Yellow', 'LightYellow', 'Brown', 'DarkYellow',
                'Gray', 'LightGray', 'DarkGray',
                'Black', 'White',
                'Orange', 'Purple', 'Violet',
            ]
        };
        var argValues = {};
        Object.keys(values).forEach(function (key) {
            var items = values[key].map(function (val) { return ({
                label: val,
                kind: vscode_languageserver_1.CompletionItemKind.EnumMember,
                documentation: '',
                insertText: val,
                insertTextFormat: vscode_languageserver_1.InsertTextFormat.PlainText
            }); });
            key.split(' ').forEach(function (name) {
                argValues[name] = items;
            });
        });
        this.highlightArgValues = argValues;
    };
    return Builtin;
}());
exports.builtinDocs = new Builtin();
//# sourceMappingURL=builtin.js.map